"""
This file is part of gempy.

Created on 21/02/2020

@author: Miguel de la Varga
"""

import numpy as np
import matplotlib.colors as mcolors
import pandas as pd

rexFileHeaderSize = 64
rexCoordSize = 22

file_header_size = 86
rexDataBlockHeaderSize = 16

file_header_and_data_header = 102
mesh_header_size = 128
all_header_size = 230

# Supported block types
# typeLineSet = 0
# typeText = 1
# typePointList = 2
typeMesh = 3
# typeImage = 4
# typeMaterial = 5
# typePeopleSimulation = 6
# typeUnityPackage = 7
# typeSceneNode = 8


n_bytes = 0


class GemPyToRex:
    def __init__(self, geo_model=None):
        """Writes GemPy data structures into binary Rex

         https://github.com/roboticeyes/openrex/blob/master/doc/rex-spec-v1.md
        """
        self.rex_bytes = bytearray()
        self.n_bytes = 0

        self.data_id = 0
        self.geo_model = geo_model

    def __call__(self, geo_model=None, meshes=True, material=True,
                 surfaces=None, topography=True, app='GemPlay'):
        """

        Args:
            meshes:
            surfaces (list): Subset of surfaces to send to the client
            app (str): Either RexViewer or GemPlay. Set of default values

        Returns:

        """

        byte_array = bytearray()
        byte_size = 0
        self.data_id = 0

        if geo_model is None:
            geo_model = self.geo_model
        else:
            self.geo_model = geo_model

        if surfaces is not None:
            raise NotImplementedError

        flip_yz, backside, vertex_color = self.default_values(app)
        # flip_yz, backside, vertex_color = False, True, False

        surface_df = self.grab_meshes(geo_model)
        if topography:
            topography_dict = self.grab_topography(geo_model)
        else:
            topography_dict = None

        # Data Blocks
        # -----------
        if material is True:
            # Material
            byte_array += self.gempy_color_to_rex_material(surface_df, topography)

        if meshes is True:
            # Mesh
            byte_array += self.gempy_meshes_to_rex(
                surface_df,
                topography_dict=topography_dict,
                flip_yz=flip_yz,
                backside=backside,
                vertex_color=vertex_color)

        # Size of all data blocks together
        byte_size += len(byte_array)

        # Write file header
        # -----------------
        n_data_blocks = self.data_id

        header_bytes = write_file_header_block(
            n_data_blocks=n_data_blocks,
            size_data_blocks=byte_size,
            start_data=file_header_size)

        return header_bytes + byte_array

    @staticmethod
    def grab_meshes(geo_model):
        """Check if surfaces are computed. And return a pandas.DataFrame with
         the meshes to be converted

        Args:
            geo_model:

        Returns:

        """

        try:
            # Drop basement
            surface_df = geo_model._surfaces.df.groupby(
                ['isActive', 'isBasement']).get_group((True, False))
        except (IndexError, KeyError):
            raise RuntimeError('No computed surfaces yet.')

        return surface_df[['surface', 'vertices', 'edges', 'color']]

    @staticmethod
    def grab_topography(geo_model):
        from scipy.spatial import Delaunay

        if geo_model._grid.topography is None or geo_model._grid.topography.values.shape[0] == 0:
            return None
        else:
            topography_dict = dict()
            topography_dict['surface'] = "Topography"
            topography_dict['vertices'] = geo_model._grid.topography.values
            # tri = Delaunay(geo_model._grid.topography.values)
            topography_dict['edges'] = Delaunay(geo_model._grid.topography.values[:, :2]).simplices
            topography_dict['color'] = geo_model.solutions.geological_map

            return topography_dict

    @staticmethod
    def hex_to_rgb(hex: str, normalize: bool = True) -> np.ndarray:
        """Transform colors from hex to rgb"""
        hex = hex.lstrip('#')
        hlen = len(hex)
        rgb = np.array([int(hex[i:i + hlen // 3], 16) for i in range(0, hlen, hlen // 3)])
        if normalize is True:
            rgb = rgb / 255
        return rgb

    @staticmethod
    def default_values(app):
        """

        Args:
            app:

        Returns:
            list: flip_yz, backside, vertex_color
        """
        if app == 'GemPlay':
            return (False, True, False)
        elif app == 'RexView':
            return (True, True, True)
        else:
            raise AttributeError('app must be either GemPlay or RexView')

    def gempy_meshes_to_rex(self,
                            surface_df,
                            topography_dict=None,
                            flip_yz=False,
                            backside=True,
                            vertex_color=False):
        """Write mesh to Rexfile.

        Args:
            surface_df:
            topography_dict:
            flip_yz: Fliping YZ coordinates. Rexview need this
            backside: If True, create a second set of triangles on the backside of the mesh
            vertex_color

        Returns:

        Notes:
            At the moment 14.07.2020 it is not possible to write normals or texture

        """
        rex_bytes = bytearray()
        mesh_number = 0

        # Loop geological surfaces surfaces
        for idx, surface_vals in surface_df.iterrows():

            tri = surface_vals['edges']
            if tri is np.nan:
                continue

            ver = surface_vals['vertices']
            surface_name = surface_vals['surface']
            if vertex_color:
                # Hex Colors
                col_ = surface_vals['color']
            else:
                col_ = None

            rex_bytes = self.mesh_prepare_and_encode(rex_bytes, mesh_number, ver, tri,
                                                     surface_name, col_=col_,
                                                     flip_yz=flip_yz, backside=backside,
                                                     vertex_color=False)
            mesh_number += 1

        # Add topography
        if topography_dict is not None:

            rex_bytes = self.mesh_prepare_and_encode(rex_bytes, n_surface=-1,
                                                     ver=topography_dict['vertices'],
                                                     tri=topography_dict['edges'],
                                                     surface_name=topography_dict['surface'],
                                                     col_=topography_dict['color'],
                                                     flip_yz=flip_yz, backside=backside,
                                                     vertex_color=True)

        return rex_bytes

    def mesh_prepare_and_encode(self, rex_bytes, n_surface, ver, tri, surface_name, col_=None, \
                                flip_yz=False,
                                backside=True,
                                vertex_color=False):

        if flip_yz:
            # This depends. For RexViewer we need to flip XYZ. For GemPlay not really
            ver_ = np.copy(ver)
            ver[:, 2] = ver_[:, 1]
            ver[:, 1] = ver_[:, 2]

        # Pre-processing GemPy output
        ver_ravel, tri_ravel, n_vtx_coord, n_triangles = mesh_preprocess(ver, tri)

        # Number of vertex colors
        if vertex_color:
            n_vtx_colors = n_vtx_coord

            # Give color to each vertex
            # TODO: Is this necessary if I pass a material
            if type(col_) is str:
                colors = np.zeros_like(ver) + self.hex_to_rgb(col_, normalize=True)
                c_r = colors.ravel()

            elif type(col_) is np.ndarray:
                surf_df = self.geo_model._surfaces.df.set_index('id')
                colors_hex = surf_df.groupby(
                    ['isActive', 'isFault']).get_group((True, False))['color']

                colors_rgb_ = colors_hex.apply(lambda val: list(mcolors.hex2color(val)))
                colors_rgb = pd.DataFrame(colors_rgb_.to_list(), index=colors_hex.index)

                sel = np.round(col_[0]).astype(int)[0]
                c_r = colors_rgb.loc[sel].values.ravel()
            else:
                raise AttributeError("col_ must be either hex string or rgb array")
        else:
            n_vtx_colors = 0
            c_r = None

        rex_bytes = self._mesh_encode(
            rex_bytes, n_surface,
            n_vtx_coord, n_triangles, n_vtx_colors,
            surface_name, ver_ravel, tri_ravel, c_r)

        if backside:
            # Coping triangles to create the backside normal of the layers
            tri_ = np.copy(tri)
            # TURN normals - One side of the normals
            tri_[:, 2] = tri[:, 1]
            tri_[:, 1] = tri[:, 2]
            # Pre-processing GemPy output
            ver_ravel, tri_ravel, n_vtx_coord, n_triangles = mesh_preprocess(ver, tri_)
            # tri = np.append(tri, tri_)

            rex_bytes = self._mesh_encode(
                rex_bytes, n_surface,
                n_vtx_coord, n_triangles, n_vtx_colors,
                surface_name, ver_ravel, tri_ravel, c_r)

        return rex_bytes

    def _mesh_encode(self, rex_bytes, material_id,
                     n_vtx_coord, n_triangles, n_vtx_colors,
                     surface_name, ver_ravel, tri_ravel, c_r):
        # Write Mesh block - header
        mesh_header_bytes = write_mesh_header(
            n_vtx_coord / 3, n_triangles / 3,
            n_vtx_colors=n_vtx_colors / 3,
            start_vtx_coord=mesh_header_size,
            start_nor_coord=mesh_header_size + n_vtx_coord * 4,
            start_tex_coord=mesh_header_size + n_vtx_coord * 4,
            start_vtx_colors=mesh_header_size + n_vtx_coord * 4,
            start_triangles=mesh_header_size +
                            ((n_vtx_coord + n_vtx_colors) * 4),
            name=surface_name,
            material_id=material_id  # self.data_id + surface_df.shape[0]
        )

        # Write Mesh block - Vertex, triangles
        mesh_block_bytes = write_mesh_coordinates(ver_ravel, tri_ravel,
                                                  colors=c_r  # When using
                                                  # material we can avoid this
                                                  )

        # Calculate the size of the mesh block
        mesh_block_size_no_data_block_header = len(mesh_header_bytes) + \
                                               len(mesh_block_bytes)  # This is cte 128

        # Write data block header for Mesh 1
        data_header_bytes = write_data_block_header(
            size_data=mesh_block_size_no_data_block_header,
            data_id=self.data_id,
            data_type=3,  # 3 for mesh
            version_data=1  # Probably useful for counting
            # the operation number
        )
        self.data_id += 1
        rex_bytes += data_header_bytes + mesh_header_bytes + mesh_block_bytes

        return rex_bytes

    def gempy_color_to_rex_material(self, surface_df, topography=False):
        rex_bytes = bytearray()

        for idx, surface_vals in surface_df.iterrows():
            # Write data block header for Material 1
            data_header_bytes = write_data_block_header(
                data_type=5,  # Material data type
                version_data=1,  # Version. Probably useful for operation counter
                size_data=68,  # Size of the block is FIXED
                data_id=self.data_id  # self.data_id
            )
            self.data_id += 1

            rgb_color = self.hex_to_rgb(surface_vals['color'], normalize=True)
            # rgb_color = [1, 1, 1]
            # Write Material
            material_bytes = write_material_data(
                ka_red=rgb_color[0], ka_green=rgb_color[1], ka_blue=rgb_color[2],
                ka_texture_ID=9223372036854775807,  # ambient
                ks_red=rgb_color[0], ks_green=rgb_color[1], ks_blue=rgb_color[2],
                ks_texture_ID=9223372036854775807,  # specular
                kd_red=rgb_color[0], kd_green=rgb_color[1], kd_blue=rgb_color[2],
                kd_texture_ID=9223372036854775807,  # diffuse
                ns=0.1,  # specular exponent
                alpha=1  # opacity
            )

            rex_bytes += data_header_bytes + material_bytes

        if topography is True:
            # Write data block header for Material 1
            data_header_bytes = write_data_block_header(
                data_type=5,  # Material data type
                version_data=1,  # Version. Probably useful for operation counter
                size_data=68,  # Size of the block is FIXED
                data_id=-1  # self.data_id
            )

            rgb_color = [1, 1, 1]
            # Write Material
            material_bytes = write_material_data(
                ka_red=rgb_color[0], ka_green=rgb_color[1], ka_blue=rgb_color[2],
                ka_texture_ID=9223372036854775807,  # ambient
                ks_red=rgb_color[0], ks_green=rgb_color[1], ks_blue=rgb_color[2],
                ks_texture_ID=9223372036854775807,  # specular
                kd_red=rgb_color[0], kd_green=rgb_color[1], kd_blue=rgb_color[2],
                kd_texture_ID=9223372036854775807,  # diffuse
                ns=0.1,  # specular exponent
                alpha=1  # opacity
            )

            self.data_id += 1
            rex_bytes += data_header_bytes + material_bytes

        return rex_bytes


def encode(input_: list):
    """Encode python objects - normally Python primitives or numpy arrays - into
    its correspondent byte representation

    Args:
        input_ (List[tuples]): List of tuples: (object, type)

    Returns:
        byte: Array of bytes
    """
    global n_bytes
    block = bytearray()

    for tup in input_:
        arr = np.array(tup[0], dtype=tup[1]).tobytes()
        n_bytes += len(arr)
        block += arr

    return block


def write_file_header_block(n_data_blocks, size_data_blocks, version=1,
                            start_data=86, srid=3876, offsets=None):
    """
    Function that writes the header block of a rexfile:

    Args:
        n_data_blocks:
        size_data_blocks:
        version (int): Version of the file
        start_data (int): Position where data start. This is after the header
         and coordinate system. If everything works fine it should be 86
        srid (int): Spatial reference system identifier (srid)
        offsets:

    Returns:

    """
    reserved = '0' * 42
    if offsets is None:
        offsets = [0, 0, 0]

    input_ = [('REX1', 'bytes'),  # REX1
              (version, 'uint16'),  # file version
              (0, 'uint32'),  # CRC32
              (n_data_blocks, 'uint16'),  # Number of DATA BLOCKS
              (start_data, 'uint16'),  # StartData
              (size_data_blocks, 'uint64'),  # Size of all data blocks
              (reserved, 'bytes'),  # Reserved
              # Coordinate system block
              (srid, 'uint32'),  # Spatial reference system identifier (srid)
              (4, 'uint16'),  # Size of the name of the used system.
              ('EPSG', 'bytes'),  # name of the used system.
              (offsets, 'float32')]  # Global x, y, z offset

    block_bytes = encode(input_)
    return block_bytes


def write_data_block_header(size_data, data_id=1, data_type=3, version_data=1):
    """Function to write a DATA BLOCK header.

    Args:
        size_data: data block size (without header)
        data_id: id which is used in the database
        data_type (int): Type of data the data block contains:
            * 0	LineSet	A list of vertices which get connected by line segments
            * 1	Text	A position information and the actual text
            * 2	PointList	A list of 3D points with color information (e.g. point cloud)
            * 3	Mesh	A triangle mesh datastructure️
            * 4	Image	A single of arbitrary format can be stored in this block
            * 5	MaterialStandard	A standard (mesh) material definition
            * 6	SceneNode	A wrapper around a data block which can be used in the scenegraph
            * 7	Track	A track is a tracked position and orientation of an AR device
        version_data: version for this data block

    Returns:

    """

    input_ = [(data_type, 'uint16'),  # data type
              (version_data, 'uint16'),  # version for this data block
              (size_data, 'uint32'),  # data block size (without header)
              (data_id, 'uint64')]  # id which is used in the database

    block_bytes = encode(input_)
    return block_bytes


def write_mesh_header(n_vtx_coord, n_triangles,
                      start_vtx_coord, start_nor_coord, start_tex_coord, start_vtx_colors,
                      start_triangles,
                      name, material_id=1,  # material_id=9223372036854775807
                      n_nor_coord=0, n_tex_coord=0, n_vtx_colors=0,
                      lod=1, max_lod=1):
    """Function to write MESH DATA BLOCK header. The header size is fixed at 128 bytes.

    Args:
        n_vtx_coord: number of vertex coordinates
        n_triangles: number of triangles
        start_vtx_coord: start vertex coordinate block (relative to mesh block start)
        start_nor_coord: start vertex normals block (relative to mesh block start)
        start_tex_coord: start of texture coordinate block (relative to mesh block start)
        start_vtx_colors: start of colors block (relative to mesh block start)
        start_triangles: start triangle block for vertices (relative to mesh block start)
        name (str): Name of the mesh
        material_id (int):  id which refers to the corresponding material block in this file
        n_nor_coord:  number of normal coordinates (can be zero)
        n_tex_coord:  number of texture coordinates (can be zero)
        n_vtx_colors: number of vertex colors (can be zero)
        lod (int): level of detail for the given geometry
        max_lod (int): maximal level of detail for given geometry

    Returns:
        bytes: array of bytes
    """

    # Strings are immutable so there is no way to modify them in place
    str_size = len(name)  # Size of the actual name of the mesh
    rest_name = ' ' * (74 - str_size)  #
    full_name = name + rest_name

    input_ = [([lod, max_lod], 'uint16'),  # Level of detail
              ([n_vtx_coord,  # number of vertex coordinates
                n_nor_coord,  # number of normal coordinates (can be zero)
                n_tex_coord,  # number of texture coordinates (can be zero)
                n_vtx_colors,  # number of vertex colors (can be zero)
                n_triangles,  # number of triangles
                start_vtx_coord,  # start vertex coordinate block (relative to mesh block start)
                start_nor_coord,  # start vertex normals block (relative to mesh block start)
                start_tex_coord,  # start of texture coordinate block (relative to mesh block start)
                start_vtx_colors,  # start of colors block (relative to mesh block start)
                start_triangles  # start triangle block for vertices (relative to mesh block start)
                ],
               'uint32'),
              (material_id, 'uint64'),
              # id which refers to the corresponding material block in this file
              (str_size, 'uint16'),  # size of the following string name
              (full_name, 'bytes')]  # name of the mesh (this is user-readable)

    block_bytes = encode(input_)
    return block_bytes


def write_mesh_coordinates(vertex, triangles, normal=None, texture=None, colors=None):
    """Block with the coordinates of a mesh. This has to go with a header!

    Args:
        vertex (numpy.ndarray[float32]): Array of vertex XYZXYZ...
        triangles (numpy.ndarray[int32]): This is a list of integers which form
         one triangle. Please make sure that normal and texture coordinates are inline with the
         vertex coordinates. One index refers to the same normal and texture position. The
         triangle orientation is required to be counter-clockwise (CCW)
        normal (numpy.ndarray):
        texture (numpy.ndarray):
        colors (numpy.ndarray):

    Returns:

    """

    ver = vertex.ravel()
    tri = triangles.ravel()
    if normal is None:
        normal = []
    if texture is None:
        texture = []
    if colors is None:
        colors = []

    input_ = [(ver, 'float32'),
              (normal, 'float32'),
              (texture, 'float32'),
              (colors, 'float32'),
              (tri, 'uint32')]

    block_bytes = encode(input_)
    return block_bytes


def write_material_data(ka_red=255.0 / 255, ka_green=255.0 / 255, ka_blue=255.0 / 255,
                        ka_texture_ID=9223372036854775807,  # ambient
                        ks_red=255.0 / 255, ks_green=255.0 / 255, ks_blue=255.0 / 255,
                        ks_texture_ID=9223372036854775807,  # specular
                        kd_red=255.0 / 255, kd_green=255.0 / 255, kd_blue=255.0 / 255,
                        kd_texture_ID=9223372036854775807,  # diffuse
                        ns=0.1,  # specular exponent
                        alpha=1  # opacity
                        ):
    """Writes a standard material definition block

    Returns: bytes (size:68) representation of the material

    """

    input_ = [(ka_red, 'float32'), (ka_green, 'float32'), (ka_blue, 'float32'),
              (ka_texture_ID, 'uint64'),
              (ks_red, 'float32'), (ks_green, 'float32'), (ks_blue, 'float32'),
              (ks_texture_ID, 'uint64'),
              (kd_red, 'float32'), (kd_green, 'float32'), (kd_blue, 'float32'),
              (kd_texture_ID, 'uint64'),
              (ns, 'float32'), (alpha, 'float32')]

    block_bytes = encode(input_)
    return block_bytes


# TODO Move to utils
def hex_to_rgb(hex):
    """Transform colors from hex to rgb"""
    hex = hex.lstrip('#')
    hlen = len(hex)
    return tuple(int(hex[i:i + hlen // 3], 16) for i in range(0, hlen, hlen // 3))


def geomodel_to_rex(geo_model, backside=True):
    """

    Args:
        geo_model (gempy.Model):
    """

    # Fixed sizes
    mesh_header_size = 128
    file_header_size = 86

    # Init dict
    rex_bytes = {}

    # Check if surfaces are computed
    try:
        # Drop basement
        surface_df = geo_model._surfaces.df.groupby(
            ['isActive', 'isBasement']).get_group((True, False))
    except (IndexError, KeyError):
        raise RuntimeError('No computed surfaces yet.')

    # Loop surfaces
    for idx, surface_vals in surface_df.iterrows():
        ver = surface_vals['vertices']
        tri = surface_vals['edges']
        if tri is np.nan:
            break

        # Grab surface color
        col = surface_vals['color']

        # Give color to each vertex
        colors = (np.zeros_like(ver) + hex_to_rgb(col)) / 255

        # This depends. For RexViewer we need to flip XYZ. For GemPlay not really
        ver_ = np.copy(ver)
        ver_[:, 2] = ver[:, 1]
        ver_[:, 1] = ver[:, 2]
        # ----------------

        # Coping triangles to create the backside normal of the layers
        tri_ = np.copy(tri)

        # Preprocessing GemPy output
        ver_ravel, tri_ravel, n_vtx_coord, n_triangles = mesh_preprocess(ver_, tri_)

        # Calculate the size of the mesh block
        if backside is True:
            n_sides = 2
        else:
            n_sides = 1

        mesh_block_size_no_data_block_header = (2 *  # Coordinates and colors
                                                n_vtx_coord + n_triangles) * 4 + \
                                               mesh_header_size  # This is cte 128

        # Size of a MATERIAL DATA BLOCK is cte
        material_block_size_no_data_block_header = 68

        # Write file header
        if backside is True:
            n_data_blocks = 3
        else:
            n_data_blocks = 2
        header_bytes = write_file_header_block(n_data_blocks=n_data_blocks,
                                               size_data_blocks=
                                               n_sides * mesh_block_size_no_data_block_header +
                                               rexDataBlockHeaderSize +
                                               material_block_size_no_data_block_header,
                                               start_data=file_header_size)

        # Write data block header for Mesh 1
        data_bytes = write_data_block_header(size_data=mesh_block_size_no_data_block_header,
                                             data_id=1, data_type=3, version_data=1)

        # Write Mesh 1 block - header
        mesh_header_bytes = write_mesh_header(n_vtx_coord / 3, n_triangles / 3,
                                              n_vtx_colors=n_vtx_coord / 3,
                                              start_vtx_coord=mesh_header_size,
                                              start_nor_coord=mesh_header_size + n_vtx_coord * 4,
                                              start_tex_coord=mesh_header_size + n_vtx_coord * 4,
                                              start_vtx_colors=mesh_header_size + n_vtx_coord * 4,
                                              start_triangles=mesh_header_size + 2 *
                                                              (n_vtx_coord * 4),
                                              name='rock1', material_id=0)

        # Write Mesh 1 block - header
        mesh_block_bytes = write_mesh_coordinates(ver_ravel, tri_ravel, colors=colors.ravel())

        if backside:
            # Write data block header for Mesh 2
            data_bytes_r = write_data_block_header(size_data=mesh_block_size_no_data_block_header,
                                                   data_id=2, data_type=3, version_data=1)

            # TURN normals - One side of the normals
            tri_[:, 2] = tri[:, 1]
            tri_[:, 1] = tri[:, 2]

            ver_ravel, tri_ravel, n_vtx_coord, n_triangles = mesh_preprocess(ver_, tri_)

            # Write Mesh 2 block - header
            mesh_header_bytes_r = write_mesh_header(n_vtx_coord / 3, n_triangles / 3,
                                                    n_vtx_colors=n_vtx_coord / 3,
                                                    start_vtx_coord=mesh_header_size,
                                                    start_nor_coord=mesh_header_size + n_vtx_coord * 4,
                                                    start_tex_coord=mesh_header_size + n_vtx_coord * 4,
                                                    start_vtx_colors=mesh_header_size + n_vtx_coord * 4,
                                                    start_triangles=mesh_header_size + 2 *
                                                                    (n_vtx_coord * 4),
                                                    name='test_a', material_id=0)

            # Write Mesh 2 block - header
            mesh_block_bytes_r = write_mesh_coordinates(ver_ravel, tri_ravel,
                                                        colors=colors.ravel())

        # Write data block header for Material 1
        material_header_bytes = write_data_block_header(data_type=5, version_data=1, size_data=68,
                                                        data_id=0)

        # Write Material 1
        material_bytes = write_material_data()

        # Putting all data together
        if backside is True:
            all_bytes = header_bytes + data_bytes + mesh_header_bytes + mesh_block_bytes + \
                        data_bytes_r + mesh_header_bytes_r + mesh_block_bytes_r + \
                        material_header_bytes + material_bytes

        else:
            all_bytes = header_bytes + data_bytes + mesh_header_bytes + mesh_block_bytes + \
                        material_header_bytes + material_bytes

        # FOR REXView Saving each surface is a rexfile
        rex_bytes[surface_vals['surface']] = all_bytes
    return rex_bytes


def mesh_preprocess(ver, tri):
    """Prepare GemPy Output to be converted to rex

    Args:
        ver (numpy.ndarray):
        tri (numpy.ndarray):

    Returns:
        list: vertices raveled, triangels ravel, n vertex, n triangles
    """

    # TODO: Remove the type transform. Technically it does nothing
    ver_ravel = ver.ravel().astype('float32')
    tri_ravel = tri.ravel().astype('int32')
    n_vtx_coord = ver_ravel.shape[0]
    n_triangles = tri_ravel.shape[0]
    return ver_ravel, tri_ravel, n_vtx_coord, n_triangles


def write_file(bytes, path: str):
    """Write to disk a rexfile from its binary format"""

    newFile = open(path + ".rex", "wb")
    newFile.write(bytes)
    return True


def write_rex(rex_bytes: dict, path='./gempy_rex'):
    file_names = []
    e = 0
    for key, value in rex_bytes.items():
        file_name = path + key
        write_file(value, file_name)
        file_names.append(file_name + '.rex')
        e += 1

    return file_names



extent = ' (numpy.ndarray[float]): [x_min, x_max, y_min, y_max, z_min, z_max]. Extent for the visualization of data ' \
         'and ' \
         'default of for the ' \
         'regular grid class.'
resolution = '(numpy.ndarray[int]): [nx, ny, nz]'

coord = '(numpy.ndarray[float, 3]): XYZ 2D array. Axis 1 is the coordinates while axis 0 is n number of input '
coord_ori = coord + 'Notice that orientations may be place anywhere in the 3D space.'

pole_vector = '(numpy.ndarray[float, 3]): 2D numpy array where axis 1 is the gradient values G_x, G_y, G_z of the pole while axis 0 is n number of' \
              ' orientations.'

orientations = '(numpy.ndarray[float, 3]): 2D numpy array where axis 1 is are orientation values [azimuth, dip, polarity] of the pole while axis 0' \
               ' is n number of orientations. --- ' \
               '*Dip* is the inclination angle of 0 to 90 degrees measured from the horizontal plane downwards. ' \
               '*Azimuth* is the dip direction defined by a 360 degrees clockwise rotation, i.e. 0 = North,' \
               ' 90 = East, 180 = South, and 270 = West.' \
               '*Polarity* defines where the upper (geologically younger) side of the orientation plane ' \
               'is and can be declared to be either normal (1) or reversed (-1).' \
               'The orientation plane is perpendicular to the gradient.'

surface_sp = '(str, Iterable[str]): list with the surface names for each input point. They must exist in the surfaces ' \
             'object linked to SurfacePoints.'

x = '(float, Iterable[float]): values or list of values for the x coordinates'
y = '(float, Iterable[float]): values or list of values for the y coordinates'
z = '(float, Iterable[float]): values or list of values for the z coordinates'
idx_sp = '(int, list, numpy.ndarray): If passed, list of indices where the function will be applied.'

file_path = 'Any valid string path is acceptable. The string could be a URL. Valid URL schemes include http, ftp, s3,' \
            ' and file. For file URLs, a host is expected. A local file could be: file://localhost/path/to/table.csv. ' \
            'If you want to pass in a path object, pandas accepts either pathlib.Path or py._path.local.LocalPath.' \
            ' By file-like object, we refer to objects with a read() method, such as a file handler (e.g. via builtin ' \
            'open function) or StringIO.'

debug = 'of debug is True the method will return the result without modify any related object'
inplace = 'if True, perform operation in-place'

centers = '(numpy.ndarray[float, 3]): XYZ array with the center of the data. This controls how much we shift the ' \
          'input coordinates'

rescaling_factor = 'Scaling factor by which all the parameters will be rescaled.'

aesara_graph_pro = 'GemPy object that contains all graph structure of aesara'

ctrl = 'List that controls what parts of the graph for each series have to be recomputed.'

weights_vector = 'Numpy array that containing the kriging weights for each input data sorted by series.'
sfai = 'Value of the scalar field at each interface. Axis 0 is each series and axis 1 contain each surface in order'
bai = '3D array with all interpolated values for a given series and at the interfaces'
mai = 'Boolean array containing the logic to combine multiple series to obtain the final model at each interface.'
vai = '2D array with the final values once the superposition of series has been carried out at each interface.'

lith_block = ' Array with the id of each layer evaluated in each point of the regular grid. '
sfm = 'Value of the scalar field at each value of the regular grid. '
bm = '3D array with all interpolated values for a given series and at each value of the regular grid. '
mm = 'Boolean array containing the logic to combine multiple series to obtain the final model at each value of the' \
     ' regular grid. '
vm = '2D array with the final values once the superposition of series has been carried out at each value of the ' \
     'regular grid.'

vertices = 'List of numpy arrays containing the XYZ coordinates of each triangle vertex.'
edges = 'List of numpy arrays containing the indices of the vertices numpy arrays that compose each individual' \
        ' triangle. '
geological_map = '2D array containing the lithologies at the surfaces. '

recompute_rf = '(bool): if True recompute the rescaling factor'

compile_aesara = 'Default true. Either if the aesara graph must be compiled or not'
aesara_optimizer = 'Type of aesara compilation. This rules the number ' \
                   'optimizations aesara performs at compilation time: fast_run will take longer \
             to compile but at run time will be faster and will consume significantly less memory. fast_compile will \
             compile faster.'

path_i = '(str): Path to the data bases of surface_points. Default os.getcwd()'
path_o = '(str): Path to the data bases of orientations. Default os.getcwd()'
surface_points_df = '(pandas.DataFrame): A pn.Dataframe object with X, Y, Z, and surface columns'
orientations_df = '(pandas.DataFrame): A pn.Dataframe object with X, Y, Z, surface columns and pole or orientation ' \
                  'columns.'

mapping_object = '(dict, :class:`pandas.DataFrame`):\n                * dict: keys are the series and values the ' \
                 'surfaces ' \
                 'belonging to that series\n\n                * pn.DataFrame: Dataframe with surfaces as index and a ' \
                 'column series with the correspondent series name\n                  of each surface '

geo_model = '(:class:`gempy.core.model.Project`): Container class of all objects that constitute a GemPy model.'

itype = "(str{'all', 'surface_points', 'orientations', 'surfaces', 'series', 'faults', 'faults_relations','additional data'}): input data type to be retrieved."

section_dict = "(dict):  'section name': ([p1_x, p1_y], [p2_x, p2_y], [xyres, zres])"



# These two lines are necessary only if GemPy is not installed
# sys.path.append("../..")

# Importing GemPy
import gempy_legacy as gp

# Importing auxiliary libraries
import numpy as np


def test_gravity(interpolator_gravity):
    geo_model = gp.create_model('2-layers')
    gp.init_data(geo_model, extent=[0, 12, -2, 2, 0, 4], resolution=[500, 1, 500])
    geo_model.add_surfaces('surface 1')
    geo_model.add_surfaces('surface 2')
    geo_model.add_surfaces('basement')
    dz = geo_model._grid.regular_grid.dz
    geo_model._surfaces.add_surfaces_values([dz, 0, 0], ['dz'])
    geo_model._surfaces.add_surfaces_values([2.6, 2.4, 3.2], ['density'])
    geo_model.add_surface_points(3, 0, 3.05, 'surface 1')
    geo_model.add_surface_points(9, 0, 3.05, 'surface 1')

    geo_model.add_surface_points(3, 0, 1.02, 'surface 2')
    geo_model.add_surface_points(9, 0, 1.02, 'surface 2')

    geo_model.add_orientations(6, 0, 4, 'surface 1', [0, 0, 1])
    device_loc = np.array([[6, 0, 4]])

    geo_model.set_centered_grid(device_loc, resolution=[10, 10, 100], radius=16000)
    geo_model.set_aesara_function(interpolator_gravity)
    geo_model._interpolator.set_aesara_shared_gravity(pos_density=2)
    print(geo_model._additional_data)
    gp.compute_model(geo_model, set_solutions=True, compute_mesh=False)
    print(geo_model.solutions.fw_gravity)
    np.testing.assert_almost_equal(geo_model.solutions.fw_gravity, np.array([-1624.1714]), decimal=4)



import pytest

import gempy_legacy.core.grid
from gempy_legacy import Surfaces

from gempy_legacy.core.data_modules.stack import Stack

import gempy_legacy as gp
import numpy as np

from gempy_legacy.core.solution import Solution
from gempy_legacy.core.xsolution import XSolution

# tests skip if subsurface is not installed
pytest.importorskip("subsurface")

@pytest.fixture(scope='module')
def a_grid():
    # Or we can init one of the default grids since the beginning by passing
    # the correspondant attributes
    grid = gempy_legacy.core.grid.Grid(extent=[0, 2000, 0, 2000, -2000, 0],
                                       resolution=[50, 50, 50])
    grid.set_active('regular')

    grid.create_custom_grid(np.arange(12).reshape(-1, 3))
    grid.set_active('custom')

    grid.create_topography()
    grid.set_active('topography')

    section_dict = {'section_SW-NE': ([250, 250], [1750, 1750], [100, 100]),
                    'section_NW-SE': ([250, 1750], [1750, 250], [100, 100])}
    grid.create_section_grid(section_dict)
    grid.set_active('sections')

    return grid


@pytest.fixture(scope='module')
def stack_eg():
    series = Stack()
    series.set_series_index(['foo', 'foo2', 'foo5', 'foo7'])
    series.add_series('foo3')
    series.delete_series('foo2')
    series.rename_series({'foo': 'boo'})
    series.reorder_series(['foo3', 'boo', 'foo7', 'foo5'])

    series.set_is_fault(['boo'])

    fr = np.zeros((4, 4))
    fr[2, 2] = True
    series.set_fault_relation(fr)

    series.add_series('foo20')

    # Mock
    series.df['isActive'] = True
    return series


@pytest.fixture(scope='module')
def surface_eg(stack_eg):
    surfaces = Surfaces(stack_eg)
    surfaces.set_surfaces_names(['foo', 'foo2', 'foo5', 'fee'])
    surfaces.add_surfaces_values([[2, 2, 2, 6], [2, 2, 1, 8]],
                                 ['val_foo', 'val2_foo'])
    return surfaces


@pytest.fixture(scope='module')
def sol_values(a_grid):
    rg_s = a_grid.values.shape[0]
    n_input = 100
    len_x = rg_s + n_input

    n_features = 5
    n_properties = 3
    # Generate random solution
    values = list()
    values_matrix = np.random.randint(0, 10, (n_properties, len_x))
    block_matrix = np.random.randint(
        0, 10, (n_features, n_properties, len_x)
    )

    fault_block = np.random.randint(40, 50, (n_features, len_x))
    weights = None
    scalar_field = np.random.randint(20, 30, (n_features, len_x))
    unknows = None
    mask_matrix = None
    fault_mask = None

    values.append(values_matrix)
    values.append(block_matrix)
    for i in [fault_block, weights, scalar_field, unknows, mask_matrix, fault_mask]:
        values.append(i)
    return values


def test_xsol(sol_values, a_grid, stack_eg, surface_eg):
    sol = XSolution(a_grid, stack=stack_eg, surfaces=surface_eg)
    sol.set_values(sol_values)
    print('\n regular', sol.s_regular_grid)
    print('\n custom', sol.s_custom_grid)
    print('\n topo', sol.s_topography)
    print('\n sect', sol.s_sections['section_SW-NE'])
    print('\n sect', sol.s_sections['section_NW-SE'])

    sol.set_values(sol_values)
    print('\n custom2', sol.s_custom_grid)


def test_xsol_unstructured(sol_values, a_grid, stack_eg, surface_eg):
    sol = XSolution(a_grid, stack=stack_eg, surfaces=surface_eg)
    sol.set_values(sol_values)
    print('\n custom', sol.s_custom_grid)


def test_xsol_structured(sol_values, a_grid, stack_eg, surface_eg):
    sol = XSolution(a_grid, stack=stack_eg, surfaces=surface_eg)
    sol.set_values(sol_values)
    print('\n custom', sol.s_regular_grid)


def test_xsol_to_disk(sol_values, a_grid, stack_eg, surface_eg, tmpdir):
    sol = XSolution(a_grid, stack=stack_eg, surfaces=surface_eg)
    sol.set_values(sol_values)
    print('\n custom', sol.s_regular_grid)
    sol.to_netcdf(tmpdir, 'bar')


@pytest.mark.skip('Test for gempy_lite')
def test_property(sol_values, a_grid, stack_eg, surface_eg):
    sol = XSolution(a_grid, stack=stack_eg, surfaces=surface_eg)
    sol.set_values(sol_values)
    print('scalar', sol.scalar_field_matrix)
    print('lith', sol.lith_block)
    print('values', sol.values_matrix)

    print('scalar_asp', sol.scalar_field_at_surface_points)


@pytest.mark.skip('Test for gempy_lite')
def test_scalar_field_matrix_property_full(model_horizontal_two_layers):
    sol_vals = gp.compute_model(model_horizontal_two_layers, set_solutions=False)
    sol = XSolution(
        model_horizontal_two_layers._grid,
        stack=model_horizontal_two_layers._stack,
        surfaces=model_horizontal_two_layers._surfaces)
    sol.set_values(sol_vals)
    print('scalar', sol.scalar_field_matrix)

    old_sol = Solution(model_horizontal_two_layers._grid,
        series=model_horizontal_two_layers._stack,
        surfaces=model_horizontal_two_layers._surfaces)
    old_sol.set_solution_to_regular_grid(sol_vals, False, None)
    print(old_sol.scalar_field_matrix)

    np.testing.assert_array_almost_equal(sol.scalar_field_matrix,
                                         old_sol.scalar_field_matrix)


def test_xsol_full(model_horizontal_two_layers):
    model_horizontal_two_layers.update_to_interpolator()
    vals = gp.compute_model(model_horizontal_two_layers, set_solutions=False)
    sol = XSolution(
        model_horizontal_two_layers._grid,
        stack=model_horizontal_two_layers._stack,
        surfaces=model_horizontal_two_layers._surfaces)
    sol.set_values(vals)
    print('\n regular', sol.s_regular_grid)
    print('\n custom', sol.s_custom_grid)
    print('\n topo', sol.s_topography)


def test_xsol_inherit(model_horizontal_two_layers):
    vals = gp.compute_model(model_horizontal_two_layers, set_solutions=True)


def test_set_meshes(model_horizontal_two_layers):
    import subsurface
    m = model_horizontal_two_layers
    vals = gp.compute_model(m, set_solutions=True)
    unstruct = m.solutions.set_meshes(m.surfaces)
    ts = subsurface.TriSurf(unstruct)
    s  = subsurface.visualization.to_pyvista_mesh(ts)
    subsurface.visualization.pv_plot([s], image_2d=True)

    print(unstruct)


def test_save_solutions(model_horizontal_two_layers, tmpdir):
    gp.save_model(model_horizontal_two_layers, solution=True, compress=False,
                  path=tmpdir)



import gempy_legacy as gp


def test_edit_1(one_fault_model_no_interp):
    model = one_fault_model_no_interp
    rpn = gp.edit(model, 'add_surfaces', surface_list=['foo'])

    assert rpn.df.iloc[-1, 0] == 'foo'
    assert model._surfaces.df.iloc[-1, 0]



from gempy_legacy.core.grid_modules.create_topography import LoadDEMArtificial, LoadDEMGDAL
import numpy as np
import skimage.transform
import matplotlib.pyplot as plt
from scipy.constants import G
from scipy import interpolate
from gempy_legacy.utils.meta import _setdoc_pro
import gempy_legacy.utils.docstring as ds
import pandas as pn


class RegularGrid:
    """
    Class with the methods and properties to manage 3D regular grids where the model will be interpolated.

    Args:
        extent (np.ndarray):  [x_min, x_max, y_min, y_max, z_min, z_max]
        resolution (np.ndarray): [nx, ny, nz]

    Attributes:
        extent (np.ndarray):  [x_min, x_max, y_min, y_max, z_min, z_max]
        resolution (np.ndarray): [nx, ny, nz]
        values (np.ndarray): XYZ coordinates
        mask_topo (np.ndarray, dtype=bool): same shape as values. Values above the topography are False
        dx (float): size of the cells on x
        dy (float): size of the cells on y
        dz (float): size of the cells on z

    """

    def __init__(self, extent=None, resolution=None, **kwargs):
        # @ formatter:off
        self.resolution = np.ones((0 , 3)   , dtype       = 'int64')
        self.extent     = np.zeros(6 , dtype = 'float64')
        self.extent_r   = np.zeros(6 , dtype = 'float64')
        self.values     = np.zeros((0, 3))
        self.values_r   = np.zeros((0, 3))
        self.mask_topo  = np.zeros((0, 3)   , dtype       = bool)
        self.x          = None
        self.y          = None
        self.z          = None
        
        if extent is not None and resolution is not None:
            self.set_regular_grid(extent, resolution)
            self.dx, self.dy, self.dz = self.get_dx_dy_dz()
        
        # @ formatter:on

    def set_coord(self, extent, resolution):
        dx = (extent[1] - extent[0]) / resolution[0]
        dy = (extent[3] - extent[2]) / resolution[1]
        dz = (extent[5] - extent[4]) / resolution[2]

        self.x = np.linspace(extent[0] + dx / 2, extent[1] - dx / 2, resolution[0], dtype="float64")
        self.y = np.linspace(extent[2] + dy / 2, extent[3] - dy / 2, resolution[1], dtype="float64")
        self.z = np.linspace(extent[4] + dz / 2, extent[5] - dz / 2, resolution[2], dtype="float64")

        return self.x, self.y, self.z

    def create_regular_grid_3d(self, extent, resolution):
        """
        Method to create a 3D regular grid where is interpolated

        Args:
            extent (list):  [x_min, x_max, y_min, y_max, z_min, z_max]
            resolution (list): [nx, ny, nz].

        Returns:
            numpy.ndarray: Unraveled 3D numpy array where every row correspond to the xyz coordinates of a regular grid

        """

        coords = self.set_coord(extent, resolution)
        g = np.meshgrid(*coords, indexing="ij")
        values = np.vstack(tuple(map(np.ravel, g))).T.astype("float64")
        return values

    def get_dx_dy_dz(self, rescale=False):
        if rescale is True:
            dx = (self.extent_r[1] - self.extent_r[0]) / self.resolution[0]
            dy = (self.extent_r[3] - self.extent_r[2]) / self.resolution[1]
            dz = (self.extent_r[5] - self.extent_r[4]) / self.resolution[2]
        else:
            dx = (self.extent[1] - self.extent[0]) / self.resolution[0]
            dy = (self.extent[3] - self.extent[2]) / self.resolution[1]
            dz = (self.extent[5] - self.extent[4]) / self.resolution[2]
        return dx, dy, dz

    def set_regular_grid(self, extent, resolution):
        """
        Set a regular grid into the values parameters for further computations
        Args:
             extent (list, np.ndarry):  [x_min, x_max, y_min, y_max, z_min, z_max]
            resolution (list, np.ndarray): [nx, ny, nz]
        """

        self.extent = np.asarray(extent, dtype='float64')
        self.resolution = np.asarray(resolution)
        self.values = self.create_regular_grid_3d(extent, resolution)
        self.length = self.values.shape[0]
        self.dx, self.dy, self.dz = self.get_dx_dy_dz()
        return self.values

    def set_topography_mask(self, topography):
        """This method takes a topography grid of the same extent as the regular
         grid and creates a mask of voxels

        Args:
            topography (:class:`gempy.core.grid_modules.topography.Topography`):

        Returns:

        """
        assert np.array_equal(topography.extent,
                              self.extent), 'The extent of' \
                                            'the topography must match to the extent of the regular grid.'

        # interpolate topography values to the regular grid
        regular_grid_topo = skimage.transform.resize(
            topography.values_2d,
            (self.resolution[0], self.resolution[1]),
            mode='constant',
            anti_aliasing=False, preserve_range=True)

        # Reshape the Z values of the regular grid to 3d
        values_3d = self.values[:, 2].reshape(self.resolution)
        if regular_grid_topo.ndim == 3:
            regular_grid_topo_z = regular_grid_topo[:, :, [2]]
        elif regular_grid_topo.ndim == 2:
            regular_grid_topo_z = regular_grid_topo
        else:
            raise ValueError()
        mask = np.greater(values_3d[:, :, :], regular_grid_topo_z)

        self.mask_topo = mask
        return self.mask_topo


class Sections:
    """
    Object that creates a grid of cross sections between two points.

    Args:
        regular_grid: Model.grid.regular_grid
        section_dict: {'section name': ([p1_x, p1_y], [p2_x, p2_y], [xyres, zres])}
    """

    def __init__(self, regular_grid=None, z_ext=None, section_dict=None):
        if regular_grid is not None:
            self.z_ext = regular_grid.extent[4:]
        else:
            self.z_ext = z_ext

        self.section_dict = section_dict
        self.names = []
        self.points = []
        self.resolution = []
        self.length = [0]
        self.dist = []
        self.df = pn.DataFrame()
        self.df['dist'] = self.dist
        self.values = []
        self.extent = None

        if section_dict is not None:
            self.set_sections(section_dict)

    def _repr_html_(self):
        return self.df.to_html()

    def __repr__(self):
        return self.df.to_string()

    def show(self):
        pass

    def set_sections(self, section_dict, regular_grid=None, z_ext=None):
        self.section_dict = section_dict
        if regular_grid is not None:
            self.z_ext = regular_grid.extent[4:]

        self.names = np.array(list(self.section_dict.keys()))

        self.get_section_params()
        self.calculate_all_distances()
        self.df = pn.DataFrame.from_dict(self.section_dict, orient='index',
                                         columns=['start', 'stop', 'resolution'])
        self.df['dist'] = self.dist

        self.compute_section_coordinates()

    def get_section_params(self):
        self.points = []
        self.resolution = []
        self.length = [0]

        for i, section in enumerate(self.names):
            points = [self.section_dict[section][0], self.section_dict[section][1]]
            assert points[0] != points[
                1], 'The start and end points of the section must not be identical.'

            self.points.append(points)
            self.resolution.append(self.section_dict[section][2])
            self.length = np.append(self.length, self.section_dict[section][2][0] *
                                    self.section_dict[section][2][1])
        self.length = np.array(self.length).cumsum()

    def calculate_all_distances(self):
        self.coordinates = np.array(self.points).ravel().reshape(-1,
                                                                 4)  # axis are x1,y1,x2,y2
        self.dist = np.sqrt(np.diff(self.coordinates[:, [0, 2]]) ** 2 + np.diff(
            self.coordinates[:, [1, 3]]) ** 2)

    @staticmethod
    def distance_2_points(p1, p2):
        return np.sqrt(np.diff((p1[0], p2[0])) ** 2 + np.diff((p1[1], p2[1])) ** 2)

    def compute_section_coordinates(self):
        for i in range(len(self.names)):
            xy = self.calculate_line_coordinates_2points(self.coordinates[i, :2],
                                                         self.coordinates[i, 2:],
                                                         self.resolution[i][0])
            zaxis = np.linspace(self.z_ext[0], self.z_ext[1], self.resolution[i][1],
                                dtype="float64")
            X, Z = np.meshgrid(xy[:, 0], zaxis, indexing='ij')
            Y, _ = np.meshgrid(xy[:, 1], zaxis, indexing='ij')
            xyz = np.vstack((X.flatten(), Y.flatten(), Z.flatten())).T
            if i == 0:
                self.values = xyz
            else:
                self.values = np.vstack((self.values, xyz))

    def generate_axis_coord(self):
        for i, name in enumerate(self.names):
            xy = self.calculate_line_coordinates_2points(
                self.coordinates[i, :2],
                self.coordinates[i, 2:],
                self.resolution[i][0]
            )
            yield name, xy

    def calculate_line_coordinates_2points(self, p1, p2, res):
        if isinstance(p1, list):
            p1 = np.array(p1)
        if isinstance(p2, list):
            p2 = np.array(p2)
        v = p2 - p1  # vector pointing from p1 to p2
        u = v / np.linalg.norm(v)  # normalize it
        distance = self.distance_2_points(p1, p2)
        steps = np.linspace(0, distance, res)
        values = p1.reshape(2, 1) + u.reshape(2, 1) * steps.ravel()
        return values.T

    def get_section_args(self, section_name: str):
        where = np.where(self.names == section_name)[0][0]
        return self.length[where], self.length[where + 1]

    def get_section_grid(self, section_name: str):
        l0, l1 = self.get_section_args(section_name)
        return self.values[l0:l1]

    @staticmethod
    def interpolate_zvals_at_xy(xy, topography, method='interp2d'):
        """
        Interpolates DEM values on a defined section

        Args:
            xy: x (EW) and y (NS) coordinates of the profile
            topography (:class:`gempy.core.grid_modules.topography.Topography`)
            method: interpolation method, 'interp2d' for cubic scipy.interpolate.interp2d
                                             'spline' for scipy.interpolate.RectBivariateSpline

        Returns:
            numpy.ndarray: z values, i.e. topography along the profile

        """

        xj = topography.values_2d[:, 0, 0]
        yj = topography.values_2d[0, :, 1]
        zj = topography.values_2d[:, :, 2]

        if method == 'interp2d':
            f = interpolate.interp2d(xj, yj, zj.T, kind='cubic')
            zi = f(xy[:, 0], xy[:, 1])
            if xy[:, 0][0] <= xy[:, 0][-1] and xy[:, 1][0] <= xy[:, 1][-1]:
                return np.diag(zi)
            else:
                return np.flipud(zi).diagonal()
        else:
            assert xy[:, 0][0] <= xy[:, 0][
                -1], 'The xy values of the first point must be smaller than second.' \
                     'Please use interp2d as method argument. Will be fixed.'
            assert xy[:, 1][0] <= xy[:, 1][
                -1], 'The xy values of the first point must be smaller than second.' \
                     'Please use interp2d as method argument. Will be fixed.'
            f = interpolate.RectBivariateSpline(xj, yj, zj)
            zi = f(xy[:, 0], xy[:, 1])
            return np.flipud(zi).diagonal()


class CustomGrid:
    """Object that contains arbitrary XYZ coordinates.

    Args:
        custom_grid (numpy.ndarray like): XYZ (in columns) of the desired coordinates

    Attributes:
        values (np.ndarray): XYZ coordinates
    """

    def __init__(self, custom_grid: np.ndarray):
        self.values = np.zeros((0, 3))
        self.set_custom_grid(custom_grid)

    def set_custom_grid(self, custom_grid: np.ndarray):
        """
        Give the coordinates of an external generated grid

        Args:
            custom_grid (numpy.ndarray like): XYZ (in columns) of the desired coordinates

        Returns:
              numpy.ndarray: Unraveled 3D numpy array where every row correspond to the xyz coordinates of a regular
               grid
        """
        custom_grid = np.atleast_2d(custom_grid)
        assert type(custom_grid) is np.ndarray and custom_grid.shape[1] == 3, \
            'The shape of new grid must be (n,3)  where n is the number of' \
            ' points of the grid'

        self.values = custom_grid
        self.length = self.values.shape[0]
        return self.values


class CenteredGrid:
    """
    Logarithmic spaced grid.
    """

    def __init__(self, centers=None, radius=None, resolution=None):
        self.grid_type = 'centered_grid'
        self.values = np.empty((0, 3))
        self.length = self.values.shape[0]
        self.resolution = resolution
        self.kernel_centers = np.empty((0, 3))
        self.kernel_dxyz_left = np.empty((0, 3))
        self.kernel_dxyz_right = np.empty((0, 3))
        self.tz = np.empty(0)

        if centers is not None and radius is not None:
            if resolution is None:
                resolution = [10, 10, 20]

            self.set_centered_grid(centers=centers, radius=radius,
                                   resolution=resolution)

    @staticmethod
    @_setdoc_pro(ds.resolution)
    def create_irregular_grid_kernel(resolution, radius):
        """
        Create an isometric grid kernel (centered at 0)

        Args:
            resolution: [s0]
            radius (float): Maximum distance of the kernel

        Returns:
            tuple: center of the voxel, left edge of each voxel (for xyz), right edge of each voxel (for xyz).
        """

        if radius is not list or radius is not np.ndarray:
            radius = np.repeat(radius, 3)

        g_ = []
        g_2 = []
        d_ = []
        for xyz in [0, 1, 2]:

            if xyz == 2:
                # Make the grid only negative for the z axis

                g_.append(np.geomspace(0.01, 1, int(resolution[xyz])))
                g_2.append(
                    (np.concatenate(([0], g_[xyz])) + 0.05) * - radius[xyz] * 1.2)
            else:
                g_.append(np.geomspace(0.01, 1, int(resolution[xyz] / 2)))
                g_2.append(
                    np.concatenate((-g_[xyz][::-1], [0], g_[xyz])) * radius[xyz])
            d_.append(np.diff(np.pad(g_2[xyz], 1, 'reflect', reflect_type='odd')))

        g = np.meshgrid(*g_2)
        d_left = np.meshgrid(d_[0][:-1] / 2, d_[1][:-1] / 2, d_[2][:-1] / 2)
        d_right = np.meshgrid(d_[0][1:] / 2, d_[1][1:] / 2, d_[2][1:] / 2)
        kernel_g = np.vstack(tuple(map(np.ravel, g))).T.astype("float64")
        kernel_d_left = np.vstack(tuple(map(np.ravel, d_left))).T.astype("float64")
        kernel_d_right = np.vstack(tuple(map(np.ravel, d_right))).T.astype("float64")

        return kernel_g, kernel_d_left, kernel_d_right

    @_setdoc_pro(ds.resolution)
    def set_centered_kernel(self, resolution, radius):
        """
        Set a centered

        Args:
            resolution: [s0]
            radius (float): Maximum distance of the kernel

        Returns:

        """
        self.kernel_centers, self.kernel_dxyz_left, self.kernel_dxyz_right = self.create_irregular_grid_kernel(
            resolution, radius)

        return self.kernel_centers

    @_setdoc_pro(ds.resolution)
    def set_centered_grid(self, centers, kernel_centers=None, **kwargs):
        """
        Main method of the class, set the XYZ values around centers using a kernel.

        Args:
            centers (np.array): XYZ array with the centers of where we want to create a grid around
            kernel_centers (Optional[np.array]): center of the voxels of a desired kernel.
            **kwargs:
                * resolution: [s0]
                * radius (float): Maximum distance of the kernel
        Returns:

        """

        self.values = np.empty((0, 3))
        centers = np.atleast_2d(centers)

        if kernel_centers is None:
            kernel_centers = self.set_centered_kernel(**kwargs)

        assert centers.shape[
                   1] == 3, 'Centers must be a numpy array that contains the coordinates XYZ'

        for i in centers:
            self.values = np.vstack((self.values, i + kernel_centers))

        self.length = self.values.shape[0]

    def set_tz_kernel(self, **kwargs):
        if self.kernel_centers.size == 0:
            self.set_centered_kernel(**kwargs)

        grid_values = self.kernel_centers

        s_gr_x = grid_values[:, 0]
        s_gr_y = grid_values[:, 1]
        s_gr_z = grid_values[:, 2]

        # getting the coordinates of the corners of the voxel...
        x_cor = np.stack((s_gr_x - self.kernel_dxyz_left[:, 0],
                          s_gr_x + self.kernel_dxyz_right[:, 0]), axis=1)
        y_cor = np.stack((s_gr_y - self.kernel_dxyz_left[:, 1],
                          s_gr_y + self.kernel_dxyz_right[:, 1]), axis=1)
        z_cor = np.stack((s_gr_z - self.kernel_dxyz_left[:, 2],
                          s_gr_z + self.kernel_dxyz_right[:, 2]), axis=1)

        # ...and prepare them for a vectorial op
        x_matrix = np.repeat(x_cor, 4, axis=1)
        y_matrix = np.tile(np.repeat(y_cor, 2, axis=1), (1, 2))
        z_matrix = np.tile(z_cor, (1, 4))

        s_r = np.sqrt(x_matrix ** 2 + y_matrix ** 2 + z_matrix ** 2)

        # This is the vector that determines the sign of the corner of the voxel
        mu = np.array([1, -1, -1, 1, -1, 1, 1, -1])

        self.tz = (
            np.sum(- 1 *
                   G *
                   mu * (
                           x_matrix * np.log(y_matrix + s_r) +
                           y_matrix * np.log(x_matrix + s_r) -
                           z_matrix * np.arctan(
                       x_matrix * y_matrix / (z_matrix * s_r))),
                   axis=1))

        return self.tz

# class Topography:
#     """
#     Object to include topography in the model.
#     """
#     def __init__(self, regular_grid):
#         self.regular_grid = regular_grid
#         self.values = np.zeros((0, 3))
#
#         self.topo = None
#         self.values_3D = np.zeros((0, 0, 0))
#         self.extent = None
#         self.resolution = None
#
#         self.type = None
#
#     def load_from_gdal(self, filepath):
#         self.topo = Load_DEM_GDAL(filepath, self.regular_grid)
#         self._create_init()
#         self._fit2model()
#         self.type = 'real'
#
#     def load_random_hills(self, **kwargs):
#         self.topo = LoadDEMArtificial(self.regular_grid, **kwargs)
#         self._create_init()
#         self._fit2model()
#         self.type = 'artificial'
#
#     def load_from_saved(self, filepath):
#         assert filepath[-4:] == '.npy', 'The file must end on .npy'
#         topo = np.load(filepath, allow_pickle=True)
#         self.values_3D = topo[0]
#         self.extent = topo[1]
#         self.resolution = topo[2]
#         self._fit2model()
#         self.type = 'real'
#
#     def _create_init(self):
#         self.values_3D = self.topo.values_3D
#         self.extent = self.topo.extent
#         self.resolution = self.topo.resolution
#
#     # These two methods makes more sense in regular grid passing a topography
#     # object.
#     def _fit2model(self):
#         self.values = np.vstack((
#             self.values_3D[:, :, 0].ravel(), self.values_3D[:, :, 1].ravel(),
#             self.values_3D[:, :, 2].ravel())).T.astype("float64")
#
#         if np.any(self.regular_grid.extent[:4] - self.extent) != 0:
#             print('obacht')
#             # todo if grid extent bigger fill missing values with nans for chloe
#             self._crop()
#
#         if np.any(self.regular_grid.resolution[:2] - self.resolution) != 0:
#             self._resize()
#         else:
#             self.values_3D_res = self.values_3D
#
#         self.regular_grid.mask_topo = self._create_grid_mask()
#
#     def _crop(self):
#         pass
#
#     def _resize(self):
#         self.values_3D_res = skimage.transform.resize(self.values_3D,
#                                                       (self.regular_grid.resolution[0], self.regular_grid.resolution[1]),
#                                                       mode='constant',
#                                                       anti_aliasing=False, preserve_range=True)
#
#     def show(self):
#         from gempy.plot.helpers import add_colorbar
#         if self.type == 'artificial':
#             fig, ax = plt.subplots()
#             CS= ax.contour(self.values_3D[:, :, 2], extent=(self.extent[:4]), colors='k', linestyles='solid')
#             ax.clabel(CS, inline=1, fontsize=10, fmt='%d')
#             CS2 = ax.contourf(self.values_3D[:, :, 2], extent=(self.extent[:4]), cmap='terrain')
#             add_colorbar(axes=ax, label='elevation [m]', cs=CS2)
#         else:
#             im = plt.imshow(np.flipud(self.values_3D[:,:,2]), extent=(self.extent[:4]))
#             add_colorbar(im=im, label='elevation [m]')
#         plt.axis('scaled')
#         plt.xlabel('X')
#         plt.ylabel('Y')
#         plt.title('Model topography')
#
#     def save(self, filepath):
#         """
#         Save the topography file in a numpy array which can be loaded later, to avoid the gdal process.
#         Args:
#             filepath (str): path where the array should be stored.
#
#         Returns:
#
#         """
#         np.save(filepath, np.array([self.values_3D, self.extent, self.resolution]))
#         print('saved')
#
#     def _create_grid_mask(self):
#         ind = self._find_indices()
#         gridz = self.regular_grid.values[:, 2].reshape(*self.regular_grid.resolution).copy()
#         for x in range(self.regular_grid.resolution[0]):
#             for y in range(self.regular_grid.resolution[1]):
#                 z = ind[x, y]
#                 gridz[x, y, z:] = 99999
#         mask = (gridz == 99999)
#         return mask# np.multiply(np.full(self.regular_grid.values.shape, True).T, mask.ravel()).T
#
#     def _find_indices(self):
#         zs = np.linspace(self.regular_grid.extent[4], self.regular_grid.extent[5], self.regular_grid.resolution[2])
#         dz = (zs[-1] - zs[0]) / len(zs)
#         return ((self.values_3D_res[:, :, 2] - zs[0]) / dz + 1).astype(int)
#
#     def interpolate_zvals_at_xy(self, xy, method='interp2d'):
#         """
#         Interpolates DEM values on a defined section
#
#         Args:
#             :param xy: x (EW) and y (NS) coordinates of the profile
#             :param method: interpolation method, 'interp2d' for cubic scipy.interpolate.interp2d
#                                              'spline' for scipy.interpolate.RectBivariateSpline
#         Returns:
#             :return: z values, i.e. topography along the profile
#         """
#         xj = self.values_3D[:, :, 0][0, :]
#         yj = self.values_3D[:, :, 1][:, 0]
#         zj = self.values_3D[:, :, 2]
#
#         if method == 'interp2d':
#             f = interpolate.interp2d(xj, yj, zj, kind='cubic')
#             zi = f(xy[:, 0], xy[:, 1])
#             if xy[:, 0][0] <= xy[:, 0][-1] and xy[:, 1][0] <= xy[:, 1][-1]:
#                 return np.diag(zi)
#             else:
#                 return np.flipud(zi).diagonal()
#         else:
#             assert xy[:, 0][0] <= xy[:, 0][-1], 'The xy values of the first point must be smaller than second.' \
#                                                'Please use interp2d as method argument. Will be fixed.'
#             assert xy[:, 1][0] <= xy[:, 1][-1], 'The xy values of the first point must be smaller than second.' \
#                                                'Please use interp2d as method argument. Will be fixed.'
#             f = interpolate.RectBivariateSpline(xj, yj, zj)
#             zi = f(xy[:, 0], xy[:, 1])
#             return np.flipud(zi).diagonal()



import numpy as np
import pandas as pn
import scipy.stats as stats
from .joyplot import joyplot

from typing import Union
import matplotlib.gridspec as gridspect

# Create cmap
from matplotlib.colors import ListedColormap
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib.cm as cmx
import seaborn as sns
from arviz.plots.jointplot import *
from arviz.plots.plot_utils import make_label, _scale_fig_size
from arviz import plot_kde, plot_dist
from arviz.plots.jointplot import _var_names
from arviz.plots.kdeplot import _fast_kde_2d
from arviz.stats import hpd
import arviz


# Seaborn style
sns.set(style="white", rc={"axes.facecolor": (0, 0, 0, 0)})

# Discrete cmap
    # seaborn palette
pal_disc = sns.cubehelix_palette(10, rot=-.25, light=.7)
pal_disc_l = sns.cubehelix_palette(10)

    # matplotlib cmap
my_cmap = ListedColormap(pal_disc)
my_cmap_l = ListedColormap(pal_disc_l)

# Continuous cmap
    # seaborn palette
pal_cont_light = sns.cubehelix_palette(250, rot=-.25, light=1)
pal_cont = sns.cubehelix_palette(250, rot=-.25, light=.7)

pal_cont_l = sns.cubehelix_palette(250)
pal_cont_l_light = sns.cubehelix_palette(250, light=1)

    # matplotlib cmap
my_cmap_full = ListedColormap(pal_cont)
my_cmap_full_light = ListedColormap(pal_cont_light)

my_cmap_full_l = ListedColormap(pal_cont_l)
my_cmap_full_l_light = ListedColormap(pal_cont_l_light)

# Default individual colors
default_red = '#DA8886'
default_blue = pal_cont[50]
default_l = pal_disc_l.as_hex()[4]


class PlotPosterior:
    def __init__(self, data: arviz.data.inference_data.InferenceData = None):
        self.data = data
        self.iteration = 1
        self.likelihood_axes = None
        self.marginal_axes = None
        self.joy = None

    def create_figure(self, marginal=True, likelihood=True, joyplot=True,
                      figsize=None, textsize=None,
                      n_samples=11):

        figsize, self.ax_labelsize, _, self.xt_labelsize, self.linewidth, _ = _scale_fig_size(figsize, textsize)
        self.fig = plt.figure(figsize=figsize, constrained_layout=False)
        gs_0 = gridspect.GridSpec(3, 6, figure=self.fig, hspace=.1)

        if marginal is True:
            # Testing
            if likelihood is False:
                self.marginal_axes = self._create_joint_axis(figure=self.fig, subplot_spec=gs_0[0:2, 0:4])
            elif likelihood is False and joyplot is False:
                self.marginal_axes = self._create_joint_axis(figure=self.fig, subplot_spec=gs_0[:, :])
    
            else:
                self.marginal_axes = self._create_joint_axis(figure=self.fig, subplot_spec=gs_0[0:2, 0:3])

        if likelihood is True:
            if marginal is False:
                self.likelihood_axes = self._create_likelihood_axis(figure=self.fig, subplot_spec=gs_0[0:2, 0:4])
            elif joyplot is False:
                self.likelihood_axes = self._create_likelihood_axis(figure=self.fig, subplot_spec=gs_0[0:2, 4:])
            else:
                self.likelihood_axes = self._create_likelihood_axis(figure=self.fig, subplot_spec=gs_0[0:1, 4:])

        if joyplot is True:
            self.n_samples = n_samples
            if marginal is False and likelihood is False:
                self.joy = self._create_joy_axis(self.fig, gs_0[:, :])
            else:
                self.joy = self._create_joy_axis(self.fig, gs_0[1:2, 4:])

    def _create_joint_axis(self, figure=None, subplot_spec=None, figsize=None, textsize=None):
        figsize, ax_labelsize, _, xt_labelsize, linewidth, _ = _scale_fig_size(figsize, textsize)
        # Instantiate figure and grid

        if figure is None:
            fig = plt.figure(figsize=figsize, constrained_layout=True)
        else:
            fig = figure

        if subplot_spec is None:
            grid = plt.GridSpec(4, 4, hspace=0.1, wspace=0.1, figure=fig)
        else:
            grid = gridspect.GridSpecFromSubplotSpec(4, 4, subplot_spec=subplot_spec)

        # Set up main plot
        self.axjoin = fig.add_subplot(grid[1:, :-1])

        # Set up top KDE
        self.ax_hist_x = fig.add_subplot(grid[0, :-1], sharex=self.axjoin)
        self.ax_hist_x.tick_params(labelleft=False, labelbottom=False)

        # Set up right KDE
        self.ax_hist_y = fig.add_subplot(grid[1:, -1], sharey=self.axjoin)
        self.ax_hist_y.tick_params(labelleft=False, labelbottom=False)
        sns.despine(left=True, bottom=True)

        return self.axjoin, self.ax_hist_x, self.ax_hist_y

    def _create_likelihood_axis(self, figure=None, subplot_spec=None, **kwargs):
        # Making the axes:
        if figure is None:
            figsize = kwargs.get('figsize', None)
            fig = plt.figure(figsize=figsize, constrained_layout=False)
        else:
            fig = figure

        if subplot_spec is None:
            grid = plt.GridSpec(1, 1, hspace=0.1, wspace=0.1, figure=fig)
        else:
            grid = gridspect.GridSpecFromSubplotSpec(1, 1, subplot_spec=subplot_spec)

        ax_like = fig.add_subplot(grid[0, 0])
        ax_like.spines['bottom'].set_position(('data', 0.0))
        ax_like.yaxis.tick_right()

        ax_like.spines['right'].set_position(('axes', 1.03))
        ax_like.spines['top'].set_color('none')
        ax_like.spines['left'].set_color('none')
        ax_like.set_xlabel('Observations')
        ax_like.set_title('Likelihood')
        return ax_like

    def _create_joy_axis(self, figure=None, subplot_spec=None, n_samples=None, overlap=.85):
        if n_samples is None:
            n_samples = self.n_samples

        grid = gridspect.GridSpecFromSubplotSpec(n_samples, 1, hspace=-overlap, subplot_spec=subplot_spec)
        ax_joy = [figure.add_subplot(grid[i, 0]) for i in range(n_samples)]
        ax_joy[0].set_title('Foo Likelihood')

        return ax_joy

    def create_color_map(self):
        cNorm = colors.Normalize(0, self.y_max_like)
        scalarMap = cmx.ScalarMappable(norm=cNorm, cmap=my_cmap_full)
        return scalarMap

    def evaluate_cmap(self, cmap, draw_mu, draw_sigma, obs: Union[float, list] = None):
        likelihood_at_observation = stats.norm.pdf(obs, loc=draw_mu, scale=draw_sigma)
        color_fill = colors.to_hex(cmap.to_rgba(np.atleast_1d(likelihood_at_observation)[0]))
        return color_fill

    def plot_marginal_posterior(self, plotters, iteration=-1, **marginal_kwargs):
        marginal_kwargs.setdefault("plot_kwargs", {})
        marginal_kwargs["plot_kwargs"]["linewidth"] = self.linewidth
        marginal_kwargs.setdefault('fill_kwargs', {})

        marginal_kwargs["plot_kwargs"].setdefault('color', default_l)
        marginal_kwargs['fill_kwargs'].setdefault('color', default_l)
        marginal_kwargs['fill_kwargs'].setdefault('alpha', .8)

        # Flatten data
        x = plotters[0][2].flatten()[:iteration]
        y = plotters[1][2].flatten()[:iteration]

        for val, ax, rotate in ((x, self.ax_hist_x, False), (y, self.ax_hist_y, True)):
            plot_dist(val, textsize=self.xt_labelsize, rotated=rotate, ax=ax, **marginal_kwargs)

    def plot_joint_posterior(self, plotters, iteration=-1, kind='kde', **joint_kwargs):

        # Set labels for axes
        x_var_name = make_label(plotters[0][0], plotters[0][1])
        y_var_name = make_label(plotters[1][0], plotters[1][1])

        self.axjoin.set_xlabel(x_var_name, fontsize=self.ax_labelsize)
        self.axjoin.set_ylabel(y_var_name, fontsize=self.ax_labelsize)
        self.axjoin.tick_params(labelsize=self.xt_labelsize)

        # Flatten data
        x = plotters[0][2].flatten()[:iteration]
        y = plotters[1][2].flatten()[:iteration]

        if kind == "scatter":
            self.axjoin.scatter(x, y, **joint_kwargs)
        elif kind == "kde":
            if False:
                gridsize = (128, 128)# if contour else (256, 256)

                density, xmin, xmax, ymin, ymax = _fast_kde_2d(x, y, gridsize=gridsize)

             #   self.axjoin.scatter(x, y, density)
                self.axjoin.imshow(density)
            else:
                if 'contour' not in joint_kwargs:
                    joint_kwargs.setdefault('contour', True) 
                fill_last = joint_kwargs.get('fill_last', False)
                
                try:
                    self.foo = plot_kde(x, y, fill_last=fill_last, ax=self.axjoin, **joint_kwargs)
                except ValueError:
                    pass
                except np.linalg.LinAlgError:
                    pass
        else:
            gridsize = joint_kwargs.get('grid_size', 'auto')
            if gridsize == "auto":
                gridsize = int(len(x) ** 0.35)
            self.axjoin.hexbin(x, y, mincnt=1, gridsize=gridsize, **joint_kwargs)
            self.axjoin._grid(False)

    def plot_trace(self, plotters, iteration, n_iterations=20):
        i_0 = np.max([0, (iteration - n_iterations)])

        theta1_val_trace = plotters[0][2].flatten()[i_0:iteration+1]
        theta2_val_trace = plotters[1][2].flatten()[i_0:iteration+1]

        theta1_val = theta1_val_trace[-1]
        theta2_val = theta2_val_trace[-1]

        # Plot point of the given iteration
        self.axjoin.plot(theta1_val, theta2_val, 'bo', ms=6, color='k')

        # Plot a trace of n_iterations
        pair_x_array = np.vstack(
            (theta1_val_trace[:-1], theta1_val_trace[1:])).T
        pair_y_array = np.vstack((theta2_val_trace[:-1], theta2_val_trace[1:])).T
        for i, pair_x in enumerate(pair_x_array):
            alpha_val = i / pair_x_array.shape[0]
            pair_y = pair_y_array[i]
            self.axjoin.plot(pair_x, pair_y, linewidth=1, alpha=alpha_val, color='k')

    def plot_marginal(self, var_names=None, data=None, iteration=-1,
                      group='both',
                      plot_trace=True, n_iterations=20,
                      kind='kde',
                      coords=None, credible_interval=.98,
                      marginal_kwargs=None, marginal_kwargs_prior=None,
                      joint_kwargs=None, joint_kwargs_prior=None):
        
        self.axjoin.clear()
        self.ax_hist_x.clear()
        self.ax_hist_y.clear()

        if data is None:
            data = self.data

        valid_kinds = ["scatter", "kde", "hexbin"]
        if kind not in valid_kinds:
            raise ValueError(
                ("Plot type {} not recognized." "Plot type must be in {}").format(kind, valid_kinds)
            )

        if coords is None:
            coords = {}

        if joint_kwargs is None:
            joint_kwargs = {}

        if marginal_kwargs is None:
            marginal_kwargs = {}

        data_0 = convert_to_dataset(data, group="posterior")
        var_names = _var_names(var_names, data_0)

        plotters = list(xarray_var_iter(get_coords(data_0, coords), var_names=var_names, combined=True))

        if len(plotters) != 2:
            raise Exception(
                "Number of variables to be plotted must 2 (you supplied {})".format(len(plotters))
            )

        if kind == 'kde':
            joint_kwargs.setdefault('contourf_kwargs', {})
            joint_kwargs.setdefault('contour_kwargs', {})
            joint_kwargs.setdefault('pcolormesh_kwargs', {})

            joint_kwargs['contourf_kwargs'].setdefault('cmap', my_cmap_full_l_light)
            joint_kwargs['contourf_kwargs'].setdefault('levels', 11)
            joint_kwargs['contourf_kwargs'].setdefault('alpha', 1)
            joint_kwargs['contour_kwargs'].setdefault('alpha', 0)

            joint_kwargs['pcolormesh_kwargs'].setdefault('cmap', my_cmap_full_l_light)
            joint_kwargs['pcolormesh_kwargs'].setdefault('alpha', 1)


        marginal_kwargs.setdefault('fill_kwargs', {})
        marginal_kwargs.setdefault("plot_kwargs", {})
        marginal_kwargs["plot_kwargs"]["linewidth"] = self.linewidth

        marginal_kwargs["plot_kwargs"].setdefault('color', default_l)
        marginal_kwargs['fill_kwargs'].setdefault('color', default_l)
        marginal_kwargs['fill_kwargs'].setdefault('alpha', .8)

        if group == 'both' or group == 'posterior':

            self.plot_joint_posterior(plotters, kind=kind, iteration=iteration, **joint_kwargs)
            self.plot_marginal_posterior(plotters, iteration=iteration, **marginal_kwargs)

        if group == 'both' or group == 'prior':
            alpha_p = .8 if group == 'prior' else .5

            if joint_kwargs_prior is None:
                joint_kwargs_prior = {}

            if marginal_kwargs_prior is None:
                marginal_kwargs_prior = {}

            marginal_kwargs_prior.setdefault('fill_kwargs', {})
            marginal_kwargs_prior.setdefault("plot_kwargs", {})
            marginal_kwargs_prior["plot_kwargs"]["linewidth"] = self.linewidth

            if kind == 'kde':

                joint_kwargs_prior.setdefault('contourf_kwargs', {})
                joint_kwargs_prior.setdefault('contour_kwargs', {})
                joint_kwargs_prior.setdefault('pcolormesh_kwargs', {})

                joint_kwargs_prior['contourf_kwargs'].setdefault('cmap', my_cmap_full_light)
                joint_kwargs_prior['contourf_kwargs'].setdefault('levels', 11)
                joint_kwargs_prior['contourf_kwargs'].setdefault('alpha', alpha_p)
                joint_kwargs_prior['contour_kwargs'].setdefault('alpha', 0)

                joint_kwargs_prior['pcolormesh_kwargs'].setdefault('cmap', my_cmap_full_light)
                joint_kwargs_prior['pcolormesh_kwargs'].setdefault('alpha', alpha_p)


            marginal_kwargs_prior["plot_kwargs"].setdefault('color', default_blue)
            marginal_kwargs_prior['fill_kwargs'].setdefault('color', default_blue)
            marginal_kwargs_prior['fill_kwargs'].setdefault('alpha', alpha_p)

            data_1 = convert_to_dataset(data, group="prior")
            plotters_prior = list(xarray_var_iter(get_coords(data_1, coords), var_names=var_names, combined=True))

            self.plot_joint_posterior(plotters_prior, kind=kind, **joint_kwargs_prior)
            self.plot_marginal_posterior(plotters_prior, **marginal_kwargs_prior)

            x_min, x_max, y_min, y_max = self.compute_hpd(plotters_prior, credible_interval=credible_interval)

        else:
            x_min, x_max, y_min, y_max = self.compute_hpd(plotters, iteration=iteration,
                                                          credible_interval=credible_interval)
        if plot_trace is True:
            self.plot_trace(plotters, iteration, n_iterations)
        
        self.axjoin.set_xlim(x_min, x_max)
        self.axjoin.set_ylim(y_min, y_max)
        self.ax_hist_x.set_xlim(self.axjoin.get_xlim())
        self.ax_hist_y.set_ylim(self.axjoin.get_ylim())

        return self.axjoin, self.ax_hist_x, self.ax_hist_y

    @staticmethod
    def compute_hpd(plotters, iteration=-1, credible_interval=.98):
        x = plotters[0][2].flatten()[:iteration]
        y = plotters[1][2].flatten()[:iteration]
        x_min, x_max = hpd(x, credible_interval=credible_interval)
        y_min, y_max = hpd(y, credible_interval=credible_interval)
        return x_min, x_max, y_min, y_max

    def set_likelihood_limits(self, val, type):
        val = np.repeat(np.atleast_1d(val), 1)

        if type == 'x_max':
            val = val + val * .2

            try:
                self._xma_list = np.append(self._xma_list[:25], val)
                self.x_max_like = np.max(self._xma_list)
            except AttributeError:
                self._xma_list = val
                self.x_max_like = np.max(self._xma_list)

        if type == 'x_min':
            val = val - val * .2

            try:
                self._xmi_list = np.append(self._xmi_list[:25], val)
                self.x_min_like = np.min(self._xmi_list)
            except AttributeError:
                self._xmi_list = val
                self.x_min_like = np.min(self._xmi_list)

        if type == 'y_max':
            try:
                self._yma_list = np.append(self._yma_list[:25], val)
                self.y_max_like = np.max(self._yma_list)
            except AttributeError:
                self._yma_list = val
                self.y_max_like = np.max(self._yma_list)

    def plot_normal_likelihood(self, mean:Union[str, float], std:Union[str, float], obs:Union[str, float],
                               data=None, iteration=-1, x_range=None, color='auto', **kwargs):
        self.likelihood_axes.clear()

        x_limits = kwargs.get('x_limits', 4)

        if data is None:
            data = self.data

        draw = data.posterior[{'chain':0, 'draw':iteration}]
        draw_mu = draw[mean] if type(mean) is str else mean
        draw_sigma = draw[std] if type(std) is str else std
        obs = data.observed_data[obs] if type(obs) is str else obs

        self.set_likelihood_limits(draw_mu + x_limits * draw_sigma, 'x_max')
        self.set_likelihood_limits(draw_mu - x_limits * draw_sigma, 'x_min')

        if x_range is not None:
            thick_min = x_range[0]
            thick_max = x_range[1]
        else:
            thick_max = self.x_max_like # draw_mu + 3 * draw_sigma
            thick_min = self.x_min_like # draw_mu - 3 * draw_sigma

        thick_vals = np.linspace(thick_min, thick_max, 100)
        observation = np.asarray(obs)

        thick_model = draw_mu
        thick_std = draw_sigma

        nor_l = stats.norm.pdf(thick_vals, loc=thick_model, scale=thick_std)
        self.set_likelihood_limits(nor_l.max(), 'y_max')

        likelihood_at_observation = stats.norm.pdf(observation, loc=thick_model, scale=thick_std)

        if color == 'auto':
            # This operations are for getting the same color in the likelihood plot as in the joy plot
            self.cmap_l = self.create_color_map()
            color_fill = self.evaluate_cmap(self.cmap_l, draw_mu, draw_sigma, obs)

        elif color is None:
            color_fill = default_l

        else:
            color_fill = color
        y_min = (nor_l.min() - nor_l.max()) * .01
        y_max = nor_l.max() + nor_l.max() * .05

        # This is the bell

        if not 'hide_bell' in kwargs:
            self.likelihood_axes.plot(thick_vals, nor_l, color='#7eb1bc', linewidth=.5)
            self.likelihood_axes.fill_between(thick_vals, nor_l, 0, color=color_fill, alpha=.8)

        if not 'hide_bell' in kwargs and not 'hide_lines' in kwargs:
            # This are the lines spawning from the observations
            self.likelihood_axes.vlines(observation, 0.000000000001, likelihood_at_observation, linestyles='dashdot',
                                        color='#DA8886', alpha=.5)

            self.likelihood_axes.hlines(likelihood_at_observation, observation, thick_max,
                                        linestyle='dashdot', color='#DA8886', alpha=.5)

        # This are the observations         
        observation_size = self.xt_labelsize * 3

        self.likelihood_axes.scatter(observation, np.zeros_like(observation), s=observation_size, c='#DA8886')
        self.likelihood_axes.set_ylim(y_min, y_max)
        self.likelihood_axes.set_xlim(thick_min, thick_max)

        # Make the axis sexy
        self.likelihood_axes.spines['bottom'].set_position(('data', 0.0))
        self.likelihood_axes.yaxis.tick_right()

        self.likelihood_axes.spines['right'].set_position(('axes', 1.03))
        self.likelihood_axes.spines['top'].set_color('none')
        self.likelihood_axes.spines['left'].set_color('none')
        self.likelihood_axes.set_xlabel('Observations')
        self.likelihood_axes.set_title('Likelihood')

        # self.likelihood_axes.set_ylim(y_min, y_max)
        self.likelihood_axes.set_xlim(thick_min, thick_max)

       # self.likelihood_axes.set_xlim(self.x_min_like, self.x_max_like)
        self.likelihood_axes.set_ylim(0, self.y_max_like)

        return self.likelihood_axes, self.cmap_l

    def plot_joy(self, var_names: tuple = None, obs: Union[str, float] = None,
                 data=None, iteration=-1, samples_size=1000, cmap='auto'):
        """

        A0rgs:
            var_names: mu and sigma of the likelihood!
            obs:
            data:
            iteration:
            samples_size:
            cmap:

        Returns:

        """
        try:
            [i.clear() for i in self.joy]
        except TypeError:
            pass

        n_iterations = self.n_samples
        iteration_label = [None for i in range(self.n_samples)]

        if data is None:
            data = self.data

        obs = data.observed_data[obs] if type(obs) is str else obs

        data = convert_to_dataset(data, group="posterior")
        coords = {}
        var_names = _var_names(var_names, data)

        plotters = list(
            xarray_var_iter(get_coords(data, coords), var_names=var_names, combined=True))

        x = plotters[0][2].flatten()
        y = plotters[1][2].flatten()

        n_data = x.shape[0]
        # This is the special case if n_samples is smaller than the number of bells to plot
        if iteration < self.n_samples / 2:
            l_0 = 0
            l_1 = int(self.n_samples)
            iteration_label[-1] = 0
            iteration_label[0] = l_1
        elif iteration > n_data - self.n_samples/2:
            l_0 = int(n_data - self.n_samples)
            l_1 = n_data
            iteration_label[-1] = l_0
            iteration_label[0] = l_1
        else:
            l_0 = int(iteration - np.round(self.n_samples / 2)) + 1
            l_1 = int(iteration + np.round(self.n_samples / 2))
            iteration_label[-1] = l_0
            iteration_label[int(np.round(self.n_samples / 2)) - 1] = iteration - 1
            iteration_label[0] = l_1

        x = x[l_0:l_1]
        y = y[l_0:l_1]

        self.set_likelihood_limits(x + 4 * y, 'x_max')
        self.set_likelihood_limits(x - 4 * y, 'x_min')
        df = pn.DataFrame()

        color = []

        for e in range(l_1-l_0):
            e = -e - 1
            num = np.random.normal(loc=x[e], scale=y[e], size=samples_size)
            name = e + (iteration - int(n_iterations / 2))
            df[name] = num

            if obs is not None:
                self.set_likelihood_limits(stats.norm.pdf(obs, loc=x[e], scale=y[e]), 'y_max')

            if cmap is None:
                color = default_blue
            else:
                if cmap == 'auto':
                    if self.likelihood_axes is None:
                        cmap = self.create_color_map()
                    else:
                        cmap = self.cmap_l

                color.append(self.evaluate_cmap(cmap, x[e], y[e], obs))

        if self.likelihood_axes is not None:
            x_range = self.likelihood_axes.get_xlim()
        else:
            x_range = (self.x_min_like, self.x_max_like)
        f, axes = joyplot(df, bw_method=1, labels=iteration_label, ax=self.joy,
                          yrot=0,
                          range_style='all', x_range=x_range,
                          color=color,
                          grid='y',
                          fade=False, last_axis=False,
                          linewidth=.1, alpha=1);

        n_axes = len(axes[:-1])
        if int(n_axes / 2) >= iteration:
            ax_sel = axes[-iteration-1]
            ax_sel.hlines(0, ax_sel.get_xlim()[0], ax_sel.get_xlim()[1], color='#DA8886', linewidth=3)
        elif iteration > n_data - int(self.n_samples/2):
            ax_sel = axes[-iteration-self.n_samples+n_data-1]
            ax_sel.hlines(0, ax_sel.get_xlim()[0], ax_sel.get_xlim()[1], color='#DA8886', linewidth=3)
        else:
            ax_sel = axes[int(n_axes / 2)]
            ax_sel.hlines(0, ax_sel.get_xlim()[0], ax_sel.get_xlim()[1], color='#DA8886', linewidth=3)

        if obs is not None:
            triangle_size = self.xt_labelsize*5

            if self.likelihood_axes is None:
                self.joy[0].scatter(obs, np.ones_like(obs) * self.joy[0].get_ylim()[1], marker='v',
                                    s=triangle_size, c='#DA8886')
            self.joy[-1].scatter(obs, np.ones_like(obs) * self.joy[-1].get_ylim()[0],
                                 marker='^', s=triangle_size, c='#DA8886')

        return axes

    def plot_posterior(self, prior_var, like_var, obs, iteration=-1,
                       marginal_kwargs=None, likelihood_kwargs=None, joy_kwargs = None):
        if marginal_kwargs is None:
            marginal_kwargs = {}
        if likelihood_kwargs is None:
            likelihood_kwargs = {}
        if joy_kwargs is None:
            joy_kwargs = {}

        self.plot_marginal(prior_var, iteration=iteration, **marginal_kwargs)
        _, cmap = self.plot_normal_likelihood(like_var[0], like_var[1], obs, iteration=iteration,
                                                   **likelihood_kwargs)
        self.plot_joy(like_var, obs=obs, iteration=iteration, **joy_kwargs)



import numpy as np
import pandas as pd
import pandas as pn

from ..surfaces import Surfaces
from ...utils.meta import _setdoc_pro, _setdoc


@_setdoc_pro(Surfaces.__doc__)
class GeometricData(object):
    """
    Parent class of the objects which containing the input parameters: surface_points and orientations. This class
     contain the common methods for both types of data sets.

    Args:
        surfaces (:class:`Surfaces`): [s0]

    Attributes:
        surfaces (:class:`Surfaces`)
        df (:class:`pn.DataFrame`): Pandas DataFrame containing all the properties of each individual data point i.e.
        surface points and orientations
    """

    def __init__(self, surfaces: Surfaces):

        self.surfaces = surfaces
        self.df = pn.DataFrame()

    def __repr__(self):
        c_ = self._columns_rend
        return self.df[c_].to_string()

    def _repr_html_(self):
        c_ = self._columns_rend
        return self.df[c_].to_html()

    def init_dependent_properties(self):
        """Set the defaults values to the columns before gets mapped with the the :class:`Surfaces` attribute. This
        method will get invoked for example when we add a new point."""

        # series
        self.df['series'] = 'Default series'
        self.df['series'] = self.df['series'].astype('category', copy=True)

        self.df['series'] = self.df['series'].cat.set_categories(self.surfaces.df['series'].cat.categories)

        # id
        self.df['id'] = np.nan

        # order_series
        self.df['order_series'] = 1
        return self

    @staticmethod
    @_setdoc(pn.read_csv.__doc__, indent=False)
    def read_data(file_path, **kwargs):
        """"""
        if 'sep' not in kwargs:
            kwargs['sep'] = ','

        table = pn.read_csv(file_path, **kwargs)
        return table

    def sort_table(self):
        """
        First we sort the dataframes by the series age. Then we set a unique number for every surface and resort
        the surfaces. All inplace
        """

        # We order the pandas table by surface (also by series in case something weird happened)
        self.df.sort_values(by=['order_series', 'id'],
                            ascending=True, kind='mergesort',
                            inplace=True)
        return self.df

    # @_setdoc_pro(Series.__doc__)
    def set_series_categories_from_series(self, series):
        """set the series categorical columns with the series index of the passed :class:`Series`

        Args:
            series (:class:`Series`): [s0]
        """
        self.df['series'] = self.df['series'].cat.set_categories(series.df.index)
        return True

    def update_series_category(self):
        """Update the series categorical columns with the series categories of the :class:`Surfaces` attribute."""
        self.df['series'] = self.df['series'].cat.set_categories(self.surfaces.df['series'].cat.categories)

        return True

    @_setdoc_pro(Surfaces.__doc__)
    def set_surface_categories_from_surfaces(self, surfaces: Surfaces):
        """set the series categorical columns with the series index of the passed :class:`Series`.

        Args:
            surfaces (:class:`Surfaces`): [s0]

        """

        self.df['surface'] = self.df['surface'].cat.set_categories(surfaces.df['surface'])
        return True

    # @_setdoc_pro(Series.__doc__)
    def map_data_from_series(self, series, attribute: str, idx=None):
        """
        Map columns from the :class:`Series` data frame to a :class:`GeometricData` data frame.

        Args:
            series (:class:`Series`): [s0]
            attribute (str): column to be mapped from the :class:`Series` to the :class:`GeometricData`.
            idx (Optional[int, list[int]): If passed, list of indices of the :class:`GeometricData` that will be mapped.

        Returns:
            :class:GeometricData
        """
        if idx is None:
            idx = self.df.index

        idx = np.atleast_1d(idx)
        if attribute in ['id', 'order_series']:
            self.df.loc[idx, attribute] = self.df['series'].map(series.df[attribute]).astype(int)

        else:
            self.df.loc[idx, attribute] = self.df['series'].map(series.df[attribute])

        if type(self.df['order_series'].dtype) is pn.CategoricalDtype:

            self.df['order_series'] = self.df['order_series'].cat.remove_unused_categories()
        return self

    @_setdoc_pro(Surfaces.__doc__)
    def map_data_from_surfaces(self, surfaces, attribute: str, idx=None):
        """
        Map columns from the :class:`Series` data frame to a :class:`GeometricData` data frame.
        Properties of surfaces: series, id, values.

        Args:
            surfaces (:class:`Surfaces`): [s0]
            attribute (str): column to be mapped from the :class:`Series` to the :class:`GeometricData`.
            idx (Optional[int, list[int]): If passed, list of indices of the :class:`GeometricData` that will be mapped.

        Returns:
            :class:GeometricData
        """

        if idx is None:
            idx = self.df.index
        idx = np.atleast_1d(idx)
        if attribute == 'series':
            if surfaces.df.loc[~surfaces.df['isBasement']]['series'].isna().sum() != 0:
                raise AttributeError('Surfaces does not have the correspondent series assigned. See'
                                     'Surfaces.map_series_from_series.')
            self.df.loc[idx, attribute] = self.df.loc[idx, 'surface'].map(surfaces.df.set_index('surface')[attribute])

        elif attribute in ['id', 'order_series']:
            self.df.loc[idx, attribute] = (self.df.loc[idx, 'surface'].map(surfaces.df.set_index('surface')[attribute])).astype(int)
        else:

            self.df.loc[idx, attribute] = self.df.loc[idx, 'surface'].map(surfaces.df.set_index('surface')[attribute])

    def _add_surface_to_list_from_new_surface_points_or_orientations(self, idx, surface: list | str):
        if type(surface) is str: surface = [surface]
        
        # Check if self.df['surface'] is a category
        if not isinstance(self.df['surface'].dtype, pd.CategoricalDtype):
            self.df['surface'] = self.df['surface'].astype('category', copy=True)
            self.df['surface'] = self.df['surface'].cat.set_categories(self.surfaces.df['surface'].values)

        # Check if elements in surface are categories in self.df['surface'] and if not add them
        # for s in surface:
        #     if s not in self.df['surface'].cat.categories:
        #         self.df['surface'] = self.df['surface'].cat.add_categories(s)

        if isinstance(idx, (np.int64, int)):
            self.df.loc[idx, 'surface'] = surface[0]
        elif type(idx) is list:
            self.df.loc[idx, 'surface'] = surface
